/* 
 *	NIVA -	fast secret key streaming cryptoalgorithm with:
 *			(i) variable key length, 
 *			(ii) preferred source message data item size and 
 *			(iii) embedded authentication routine
 *
 *	This program uses 8-bit source message data item (i.e. source is a stream of bytes) and
 *	targeted do demonstrate my algorithm for file en-/decryption
 *	
 *	(C) Dr. Nikolay P. Vasilyev, 2005
 *
 * ATTENTION - this program can be used for algorithm evaluation only!
 */

#include "ecrypt-sc.h"
#include <string.h>
#include <stdio.h>
#include <time.h>
#define BLOCKSIZE 16384 /* number of bytes to be en-/decrypted at once */

/* global variables */

u8 X0[8], X[8], Y0[8], Y[8]; /* core variables. X - main, Y - additional 
			      * Vars with 0-suffix reflects a previous state */

u8 A[256][8], B[256][8]; /* binary quotients for core variables, Ai for X, Bj for Y */

u8 M, N, O, P, Q, R; /* not enough space to describe these vars here - look at my paper :) */
u32 streamPointer=0; /* global source message byte counter */
int k=-1,m=-1, maxPower=0,maxPowerY=0; /* needed to hold state of previous block ciphering */

/* end of globvars definition */

int nivaEncrypt(u8* dstBlock, u8* srcBlock, int size)
{
	int i,j,l,n,o,p,r;
	
	for(i=0;i<size;i++)
	{
		/* main phase */

		if (streamPointer%O==0)
		{
			k=(k+1)%M;
			for(j=0;j<8;j++)
			{
				if (A[k][j]==1)
					maxPower=j;
			}
		}
		X[0]=X0[maxPower];
		for(l=1;l<=maxPower;l++)
		{
			if (A[k][maxPower-l+1]==1)
				X[l]=X0[l-1]+X0[maxPower];
			else
				X[l]=X0[l-1];
		}
		for(l=0;l<=maxPower;l++)
			X0[l]=X[l];

		/* deviation phase */
		if (streamPointer%P==0) 
		{
			m=(m+1)%N;
			for(n=0;n<8;n++)
			{
				if (B[m][n]==1)
					maxPowerY=n;
			}

			Y[0]=Y0[maxPowerY];
			for(o=1;o<=maxPowerY;o++)
			{
				if (B[m][maxPowerY-o+1]==1)
					Y[o]=Y0[o-1]+Y0[maxPowerY];
				else
					Y[o]=Y0[o-1];
			}
			for(o=0;o<=maxPowerY;o++)
				Y0[o]=Y[o];
			p=Y[Q%8]%8;
			r=Y[R%8]%8;
			X0[p]=(X0[p]<<r)|(X0[p]>>(8-r));
		}
		
		/* exactly encryption */
		X[0]=srcBlock[i]+X[0];
		dstBlock[i]=X[0];
		streamPointer++;
		/* printf("Decrypted byte = %d, encrypted byte = %d\n", srcBlock[i],dstBlock[i]); */
	}
		
	return 0;
}

int nivaDecrypt(u8* dstBlock, u8* srcBlock, int size)
{
	int i,j,l,n,o,p,r;
	
	for(i=0;i<size;i++)
	{
		/* main phase */
		if (streamPointer%O==0)
		{
			k=(k+1)%M;
			for(j=0;j<8;j++)
			{
				if (A[k][j]==1)
					maxPower=j;
			}
		}
		X[0]=X0[maxPower];
		for(l=1;l<=maxPower;l++)
		{
			if (A[k][maxPower-l+1]==1)
				X[l]=X0[l-1]+X0[maxPower];
			else
				X[l]=X0[l-1];
		}
		for(l=0;l<=maxPower;l++)
			X0[l]=X[l];

		/* deviation phase */
		if (streamPointer%P==0) 
		{
			m=(m+1)%N;
			for(n=0;n<8;n++)
			{
				if (B[m][n]==1)
					maxPowerY=n;
			}
		
			Y[0]=Y0[maxPowerY];
			for(o=1;o<=maxPowerY;o++)
			{
				if (B[m][maxPowerY-o+1]==1)
					Y[o]=Y0[o-1]+Y0[maxPowerY];
				else
					Y[o]=Y0[o-1];
			}
			for(o=0;o<=maxPowerY;o++)
				Y0[o]=Y[o];
			p=Y[Q%8]%8;
			r=Y[R%8]%8;
			X0[p]=(X0[p]<<r)|(X0[p]>>(8-r));
		
		}
		/* exactly decryption */
		dstBlock[i]=srcBlock[i]-X[0];
		streamPointer++;
		/* printf("Encrypted byte = %d, decrypted byte = %d\n", srcBlock[i],dstBlock[i]);*/
	}
		
	return 0;

}

int parseKey(u8* key, int keySize) /* splits a key string to values used in ciphering */
{
/*
 * The key consists of these fields:
 * 0-7 bytes: initial state of X0 variables
 * 8-15		: initial state of Y0 variables
 * 16		: value of M
 * 17		: quotients A[0][0 - 7]
 *	...
 * 17+M-1	: quotients A[M-1][0 - 7]
 * 17+M		: value of N
 * 17+M+1	: quotients B[0][0 - 7]
 *	...
 * 17+M+N	: quotients B[N-1][0 - 7]
 * 17+M+N+1	: value of O
 * 17+M+N+2	: value of P
 * 17+M+N+3	: value of Q
 * 17+M+N+4	: value of R
 *
 * Assuming that the max value for M, N is 255, max key length is 531 byte, 
 * i.e. more than 4K bits! 
 *
 * Moreover, the algorithm is general and everyone can write a program with much more larger keys
 * But, we cannot produce a mirracle, and longer keys will reduce ciphering perfomance.
 *
 * If you want, you can see the parsed key in the file "parsed.key". To achieve this uncomment fprintf etc. functions below
 */

	/* FILE* logStream=fopen("parsed.key","wb"); */
	
	if (keySize<19)
	{
		printf("Key is too short - it must be at least 19 bytes (152-bit)...\n");
		return 101;
	}

	int i,j; 
	u8 tmp; 
	
	/* initial state for X0 */
	for(i=0;i<8;i++)
	{
		X0[i]=key[i];
	/*	fprintf(logStream,"X0[%d]=%d\n",i,X0[i]); */
	}
	/* initial state for Y0 */
	for(i=8;i<16;i++)
	{
		Y0[i-8]=key[i];
	/*	fprintf(logStream,"Y0[%d]=%d\n",i-8,Y0[i-8]);*/
	}
	/* Quotients A retrieving */
	M=key[16];
	/*fprintf(logStream,"M=%d\n",M);*/
	if (keySize<=16+M)
	{
		printf("Incorrect key value...\n");
		return 103;
	}
	for(i=17;i<17+M;i++)
	{
		tmp=key[i];
		for(j=0;j<8;j++)
		{
			A[i-17][j]=tmp%2;
			tmp>>=1;
			/*fprintf(logStream,"A[%d][%d]=%d\n",i-17,j,A[i-17][j]);*/
		}
	}

	/* Quotients B retrieving */
	N=key[17+M];
	/*fprintf(logStream,"N=%d\n",N);*/
	if (N==0) /* min key length case */
		return 0; /* stop parsing */
	if (keySize<=17+M+N+4)
	{
		printf("Incorrect key value...\n");
		return 104;
	}

	for(i=17+M+1;i<=17+M+N;i++)
	{
		tmp=key[i];
		for(j=0;j<8;j++)
		{
			B[i-(17+M+1)][j]=tmp%2;
			tmp>>=1;
			/*fprintf(logStream,"B[%d][%d]=%d\n",i-(17+M+1),j,B[i-(17+M+1)][j]);*/
		}
	}
	/* special vars */
	O=key[17+M+N+1];
	/*fprintf(logStream,"O=%d\n",O);*/
	P=key[17+M+N+2];
	/*fprintf(logStream,"P=%d\n",P);*/
	Q=key[17+M+N+3];
	/*fprintf(logStream,"Q=%d\n",Q);*/
	R=key[17+M+N+4];
	/*fprintf(logStream,"R=%d\n",R);
	fclose(logStream);*/
	printf("Effective key length is %d bytes (%d bits)\n",17+M+N+4,(17+M+N+4)*8);
	/* Encryption all the key symbols to take them into opinion can be made. 
	 * This operation will effect only for internal variables
	 * If you don't wish to do it - comment 2 rows below */
	/*
		u8 pseudoResult[5120];	
		int pseudeEncryptResult=nivaEncrypt(pseudoResult, (u8*)key, keySize);
	 */

	return 0;
}

void showHelp()
{
	printf("\nConsole file encryption utility with a variable key length.\n");
	printf("\t(C) Dr. N.P. Vasilyev, 2005\n");
	printf("\nSyntax:\nniva8 {/e | /d} {/key=<key_sequence> | /keyfile=<key file name>} <source file name> <destination file name> [<message digest file name>]\n");
	printf("\nParameters:\n/e for encryption, /d for decryption\n");
	printf("/key immediately defines the key, while ");
	printf("/keyfile points to the file containing a key sequence for the ciphering operation\n");
	printf("<source file name> - the source for cipher operation\n");
	printf("<destination file name> - the result of encryption/decryption\n");
	printf("If <message digest file name> is given that file will be formed during encryption and used during decryption for digest checking.\n");
	printf("If digest checking fails, there is a decryption error (different cryptokeys, source file integrity error, etc.\n");
}

int main(int argc, char* argv[])
{
	/* checking for arguments */
	if ((argc<5)||(argc>6))
	{
		showHelp();
		return 1;
	}
	
	/* key analysis */
	char* keyParameter=strstr(argv[2],"/key=");
	char* keyFileName=strstr(argv[2],"/keyfile=");
	if (keyParameter!=NULL)
	{
		/* key data is given immediately */
		keyParameter+=5; /* bypassing /key= */
		if (parseKey((u8*)keyParameter, strlen(keyParameter))!=0) 
		{
			printf("Key parse error...\n");
			return 10;
		}
	}
	else if (keyFileName!=NULL)
	{
		/* key data is given in a file */
		keyFileName+=9;
		FILE* keyStream=fopen(keyFileName,"rb");
		if (keyStream==NULL)
		{
			printf("Key file opening error...\n");
			return 11;
		}
		char key[5120];
		int keyLength=fread(key,1,5120,keyStream);
		fclose(keyStream);
		if (keyLength<=0)
		{
			printf("Key file read error...\n");
			return 12;
		}

		/* key processing */
		int parseKeyResult=parseKey((u8*)key, keyLength);
		if (parseKeyResult!=0)
		{
			printf("Key file parse error...\n");
			return parseKeyResult;
		}
	}
	else
	{
		printf("Incorrect key definition parameter\n");
		showHelp();
		return(2);
	}

	/* checking files */
	FILE*	srcStream=fopen(argv[3],"rb");
	if (srcStream==NULL)
	{
		printf("Source file opening error...\n");
		return 3;
	}
	FILE*	dstStream=fopen(argv[4],"wb");
	if (dstStream==NULL)
	{
		printf("Destination file creation error...\n");
		return 4;
	}

	int		bytesRead=0,bytesWritten=0;
	u8		srcBuffer[BLOCKSIZE],dstBuffer[BLOCKSIZE];
	/* time calculations */
	time_t t1,t2;
	double resTime;
	time(&t1);	
	if (strcmp(argv[1],"/e")==0)
	{/* encryption */
		printf("Encryption");
		do
		{	bytesRead=fread(srcBuffer,1,BLOCKSIZE,srcStream);
			if (nivaEncrypt(dstBuffer,srcBuffer,bytesRead)!=0)
			{
				printf("There was an error during encryption...\n");
				return 21;
			}
			bytesWritten=fwrite(dstBuffer,1,bytesRead,dstStream);
			printf(".");
		} while (bytesRead>0);
		fclose(srcStream);
		fclose(dstStream);
		printf("Done\n");
	/* 
	 * MAC (the same as digest or digital file signature) calculating 
	 */
		if (argc==6)
		{
			char*  MACFileName=argv[5];
			FILE*	MACStream;
			MACStream=fopen(MACFileName,"wb");
			if (MACStream==NULL)
			{
				printf("MAC file creating error...\n");
				return 30;
			}
			bytesWritten=fwrite(X0,1,8,MACStream);
			fclose(MACStream);
			printf("Message Authentication Code (MAC) has been stored in file %s\n",MACFileName);
		}
	}
	else if (strcmp(argv[1],"/d")==0)
	{/* decryption */
		printf("Decryption");
		do
		{	bytesRead=fread(srcBuffer,1,BLOCKSIZE,srcStream);
			if (nivaDecrypt(dstBuffer,srcBuffer,bytesRead)!=0)
			{
				printf("There was an error during decryption...\n");
				return 22;
			}
			bytesWritten=fwrite(dstBuffer,1,bytesRead,dstStream);
			printf(".");
		} while (bytesRead>0);
		fclose(srcStream);
		fclose(dstStream);
		printf("Done\n");

	/* 
	 * MAC (the same as message digest or  file's digital signature) checking
	 */
 		if (argc==6)
		{
			char*	MACFileName=argv[5];
			FILE*	MACStream;
			int		diffX=0;
			MACStream=fopen(MACFileName,"rb");
			if (MACStream==NULL)
			{
				printf("MAC file opening error...\n");
				return 40;
			}
			u8 MAC[8];
			printf("Message Authentication Code (MAC) checking...");
			bytesRead=fread(MAC,1,8,MACStream);
			fclose(MACStream);
			int s;
			for(s=0;s<8;s++)
				if (X0[s]!=MAC[s])
					diffX++;			
			
			if (diffX!=0)
				printf("FAILED!\nI don't recommend you to use this decrypted file (%s)!\n",argv[4]);
			else
				printf("OK!\n");
		}
	}
	else
	{
		printf("Incorrect first parameter...\n");
		showHelp();
		return 5;
	}
	printf("Operation complete.\n");
	
	time(&t2);
	resTime=difftime(t2,t1);
	if (resTime>0)
		printf("Average perfrormance: %f Bytes/sec\n",streamPointer/resTime);
	
	return 0;
}